function [y, Pyy, Pxy] = ut(x, Pxx, g, alpha, beta, kappa)
% Scaled unscented transform
%
%   Technika umoujc aproximovat stedn hodnotu a kovariann matici
%   veliiny Y, kter vznikne transformac veliiny X nelinern funkc, 
%   Y = g(X)
%
%
%   [y, Pyy] = ut(x, Pxx, g, alpha, beta, kappa)
%   [y, Pyy, Pxy] = ut(x, Pxx, g, alpha, beta, kappa)
%
%   Vstupy: 
%   x ....... stedn hodnota veliiny X
%   Pxx ..... kovariann matice veliiny X
%   g ....... obecn nelinern transforman funkce
%   alpha, 
%   beta, 
%   kappa ... parametry ovlivujc vbr sigma bod a stanoven jejich vah
%   
%   Vstupy: 
%   y ....... aproximace stedn hodnoty veliiny Y
%   Pyy ..... aproximace kovariann matice veliiny Y
%   Pxy ..... aproximace vzjemn kovariann matice veliin X a Y
%
%
%   Datum posledn pravy: 3. 3. 2018, KF


% Inicializan st
n = length(x);
p = 2 * n + 1;
Chi = zeros(n, p);
x = x(:);

lambda = alpha^2 * (n + kappa) - n;

[U, S, ~] = svd(Pxx);
Cxx = U * sqrt(S) * U';

% Vpoet sigma bod
Chi(:, 1) = x;
c = sqrt(n + lambda);
for i = 1 : n
    Chi(:, i + 1) = x + c * Cxx(:, i);
    Chi(:, n + i + 1) = x - c * Cxx(:, i);
end

% Stanoven pslunch vah bod
w0M = lambda / (n + lambda);
w0C = w0M + 1 - alpha^2 + beta;
w = 1 / (2 * (n + lambda));

% Transformace bod funkc g
for i = 1 : p
    Xi(:, i) = g(Chi(:, i));
end

% Vpoet stedn hodnoty a kovariann matice transformovan veliiny
y = w0M * Xi(:, 1) + w * sum(Xi(:, 2 : p), 2);
Pyy = w0C * (Xi(:, 1) - y) * (Xi(:, 1) - y)';
for i = 2 : p
    Pyy = Pyy + w * (Xi(:, i) - y) * (Xi(:, i) - y)';
end

% Vpoet vzjemn kovariann matice veliin X a Y
if nargout == 3
    Pxy = w0C * (Chi(:, 1) - x) * (Xi(:, 1) - y)';
    for i = 2 : p
        Pxy = Pxy + w * (Chi(:, i) - x) * (Xi(:, i) - y)';
    end
end

end

