% Rekonstrukce rozmazanho obrazu pomoc algoritmu reduced update Kalman filter 
% (steady-state approximate RUKF)
%
% dle Woods, Ingle: Kalman filtering in two dimensions: further results

clear

% Naten pvodnho idelnho obrazu
X = imread('HK.jpg');
X = 0.2989 * X(:, :, 1) + 0.5870 * X(:, :, 2) + 0.1140 * X(:, :, 3);
X = double(X) ./ double(intmax(class(X)));
[s1, s2] = size(X);

figure
imshow(X, [0 1])

% Parametry modelu obrazu
n = 3;
c = [0.020380445 -0.002564352 -0.000146032
     0.009573138  0.048825988  0.031307254
     0.148352208  0.738287199  0];

% Rozmazn obrazu 
% (simulace rozmazn vzniklho pohybem v horizontlnm smru)
m = 7;
h = 1/m * ones(1, m);
XX = extendimage(X, m);
Z = zeros(s1, s2);
for i = 1 : s1
    for j = 1 : s2
        Z(i, j) = computesum(XX(i+m, j+1 : j+m), h);
    end
end

% Pidn gaussovskho blho umu do rozmazanho obrazu
Z = Z + randn(s1, s2) ./ 1000;

figure
imshow(Z, [0 1])

% Centrovn obrazu
imgmean = mean(Z(:));
Z = Z - imgmean;

% Rozptyl systmovho umu a umu men
q = 1e-4;
r = 1e-6;

% Velikosti oblast (okol zpracovvanho pixelu), kde je algoritmus aplikovn
% (hodnoty lze volit experimentln)
u1 = m;
u2 = m;
u3 = m;
t = u3 + m;
ZZ = extendimage(Z, t);

Xest = ZZ;
P = 0.1 * ones(s1+2*t, s2+2*t, u1, u3);

% Zjitn ustlen hodnoty Kalmanova zisku (steady-state Kalman gain)
% (zde pro 100 iterac algoritmu)
K = zeros(u1, u3);
for i = 1 + t : 1 + t
    for j = 1 + t : 100 + t
        
        % Apriorn chybov kovariann "matice"
        Ptmp = P;
        for k = 1 : u1
            for l = 1 : u3
                Ptmp(i, j, k, l) = computesum(P(i-n+1 : i, j-n+1 : j, k, l), c);
            end
        end 
        Ptmp(i, j, end, end) = computesum(squeeze(P(i, j, end-n+1 : end, end-n+1 : end)), c) + q;
        P = Ptmp; 
        
        % Kalmanv zisk
        K = zeros(u1, u3);
        k2 = 0;
        for l = 1 : m
            k2 = k2 + h(l) * computesum(squeeze(P(i, j-m+l, end, end-m+1 : end)), h);
        end
        k2 = k2 + r;
        for k = 1 : u1
            for l = 1 : u3
                k1 = computesum(P(i, j-m+1 : j, k, l), h);
                K(k, l) = k1 / k2;
            end
        end
        
        % Aposteriorn chybov kovariann "matice"
        Ptmp = P;
        for k = 1 : u1
            for l = 1 : u3
                ii = i-u1+k;
                jj = j-u3+l;
                for o = 1 : u1
                    for p = 1 : u3
                        Ptmp(ii, jj, o, p) = P(ii, jj, o, p) - K(k, l) * computesum(P(ii, jj-m+1 : jj, o, p), h);
                    end
                end
            end
        end
        P = Ptmp;
        
    end
end

% Vlastn aplikovn RUKF
for i = 1 + t : s1 + t
    for j = 1 + t : s2 + t
        
        % Apriorn odhad (predikce)
        Xest(i, j) = computesum(Xest(i-n+1 : i, j-n+1 : j), c);
        
        % Aposteriorn odhad (korekce)
        Xtmp = Xest;
        for k = 1 : u1
            for l = 1 : u2
                ii = i-u1+k;
                jj = j-u2+l;
                Xtmp(ii, jj) = Xest(ii, jj) + K(k, l+u3-u2) * (ZZ(ii, jj) - computesum(Xest(ii, jj-m+1 : jj), h));
            end
        end
        Xest = Xtmp;
        
    end
end

% Vsledn rekonstruovan obraz
Xestimated = Xest(1+t : s1+t, 1+t : s2+t) + imgmean;

figure
imshow(Xestimated, [0 1])


function [S] = computesum(A, B)
    AB = A .* B;
    S = sum(AB(:));
end

