% Algoritmus 3
%
% Jedn se o lohu statick lokalizace, piem jako men jsou chpny
% namen sly signl z jednotlivch vyslacch zazen (beacon)
%
% Pedpokldaj se datov soubory s umstnm 'data\exp3\exp3_X.csv', 
% kde X = 1, 2, ..., 20 je slo testovacho bodu, 
% a soubor 'data\exp3\calibrated.csv' reprezentujc referenn mapu
% fingerprint

clear

filename = 'data\exp3\exp3_1.csv';

calibratedRead = csvread('data\exp3\calibrated.csv');
positionTrue = csvread(filename, 0, 0, [0 0 0 1]);
measurements = csvread(filename, 1, 0);

n = size(calibratedRead, 1) / 3;
m = size(calibratedRead, 2);
calibrated(1 : n) = struct('position', zeros(1, 2), 'means', zeros(1, m), 'variances', zeros(1, m));
for i = 1 : n
    calibrated(i).position = calibratedRead(3 * (i - 1) + 1, 1 : 2)';
    calibrated(i).means = calibratedRead(3 * (i - 1) + 2, :)';
    calibrated(i).variances = calibratedRead(3 * (i - 1) + 3, :)';
end

Phi = [1 0; 0 1];
Q = [100 0; 0 100];

diffMin = Inf;
for i = 1 : n
    diff = sum(abs(measurements(1, :)' - calibrated(i).means));
    if diff < diffMin
        diffMin = diff;
        x0 = calibrated(i).position;
    end
end

x = x0;
P = [10000 0; 0 10000];

Z = measurements';

f = size(measurements, 1);
est(1 : f) = struct('x', zeros(2, 1), 'P', zeros(2, 2));

% Fingerprint Kalman filter
for k = 1 : f

    x = Phi * x;
    P = Phi * P * Phi' + Q;
        
    beta = zeros(n, 1);
    pEst = zeros(2, 1);
    zEst = zeros(m, 1);
    Pxx = zeros(2, 2);
    Pxz = zeros(2, m);
    Pzz = zeros(m, m);
    Pp = diag([(50 ^ 2) / 12 (50 ^ 2) / 12]);
    
    sumPdf = 0;
    for i = 1 : n
        sumPdf = sumPdf + pdfnorm(calibrated(i).position, x, P);
    end
    
    for i = 1 : n
        beta(i) = pdfnorm(calibrated(i).position, x, P) / sumPdf;
        pEst = pEst + beta(i) * calibrated(i).position;
        zEst = zEst + beta(i) * calibrated(i).means;
    end
    
    for i = 1 : n
        Pxx = Pxx + beta(i) * (Pp + (calibrated(i).position - pEst) * (calibrated(i).position - pEst)');
        Pxz = Pxz + beta(i) * (calibrated(i).position - pEst) * (calibrated(i).means - zEst)';
        Pzz = Pzz + beta(i) * (diag(calibrated(i).variances) + (calibrated(i).means - zEst) * (calibrated(i).means - zEst)');
    end
    
    x = pEst + Pxz * inv(Pzz) * (Z(:, k) - zEst);
    P = Pxx - Pxz * inv(Pzz) * Pxz';
    
    est(k).x = x;
    est(k).P = P;
end

positionEst = [est.x]';

k = 1 : f;
positionTrueRep = repmat(positionTrue, f, 1);

e_xhat = sqrt(sum((positionTrueRep - positionEst).^2, 2));

figure
subplot(2, 2, 1) 
plot(k, positionTrueRep(:, 1), 'g-o', k, positionEst(:, 1), 'b-*')
xlabel('as'), ylabel('Poloha (x)')
subplot(2, 2, 2) 
plot(k, positionTrueRep(:, 2), 'g-o', k, positionEst(:, 2), 'b-*')
xlabel('as'), ylabel('Poloha (y)')
legend({'Skuten hodnota', 'Hodnota odhadnut pomoc KF'}, 'Location', 'southeast')
subplot(2, 2, [3, 4]) 
plot(k, e_xhat, 'b-*', k, 0, 'k')
xlabel('as'), ylabel('Chyba')
legend({'Chyba odhadu'}, 'Location', 'northeast')

