% Ukzka pouit Kalmanova filtru pro predikovn, filtrovn a vyrovnvn
% v asov ad reprezentovan ARMA(2,1) modelem

clear

y = csvread('data_ARMA(2,1).csv');
n = length(y);
phi_1 = 0.6;
phi_2 = 0.2;
theta_1 = 0.1;
sigma2_e = 10;

sys.Phi = [phi_1 1; phi_2 0];
sys.H = [1 0];
sys.G = [1; theta_1];
sys.Q = sigma2_e;
sys.R = sigma2_e;

% Predikovn
y_predicted = zeros(1, n+5);
y_predicted(1) = y(1);
y_predicted(2) = y(2);
e = zeros(1, n);
for t = 3 : n % jednokrokov predikce ve sledovanm obdob
    x = sys.Phi * [y(t-1); phi_2 * y(t-2) + theta_1 * e(t-1)];
    y_predicted(t) = x(1);
    e(t) = y(t) - y_predicted(t);
end
x = [y(n); phi_2 * y(n-1) + theta_1 * e(n-1)];
for t = n+1 : n+5 % predikce pro 5 budoucch asovch okamik
    x = sys.Phi * x;
    y_predicted(t) = x(1);
end

% Filtrovn
% (ekvivalentn lze pout funkci kf(sys, init, y))
est_f(1 : n) = struct('x_', zeros(2, 1), 'P_', zeros(2, 2),...
    'x', zeros(2, 1), 'P', zeros(2, 2));
est_f(1).x = [y(1); 1];
est_f(1).P = [10 0; 0 10];
for t = 2 : n 
    est_f(t).x_ = sys.Phi * est_f(t-1).x;
    est_f(t).P_ = sys.Phi * est_f(t-1).P * sys.Phi' + sys.G * sys.Q * sys.G';
    K = est_f(t).P_ * sys.H' * inv(sys.H * est_f(t).P_ * sys.H' + sys.R);
    est_f(t).x = est_f(t).x_ + K * (y(t) - sys.H * est_f(t).x_);
    est_f(t).P = est_f(t).P_ - K * sys.H * est_f(t).P_;
end
estx = [est_f.x];
y_filtered = estx(1, :);

% Vyrovnvn (RauchTungStriebel smoother)
est_s(1 : n) = struct('x', zeros(2, 1), 'P', zeros(2, 2));
est_s(n).x = est_f(n).x;
est_s(n).P = est_f(n).P;
for t = n-1 : -1 : 2
    A = est_f(t).P * sys.Phi' * inv(est_f(t+1).P_);
    est_s(t).x = est_f(t).x + A * (est_s(t+1).x - est_f(t+1).x_);
    est_s(t).P = est_f(t).P + A * (est_s(t+1).P - est_f(t+1).P_) * A';
end
estx = [est_s.x];
y_smoothed = estx(1, :);
y_smoothed(1) = y(1);

t = 1 : n;
figure
subplot(3, 1, 1) 
p = plot(t, y, 'k', 1 : n+5, y_predicted, 'b');
p(2).LineWidth = 1.5;
xlim([0 105])
title('Predikovn')
subplot(3, 1, 2)
p = plot(t, y, 'k', t, y_filtered, 'b');
p(2).LineWidth = 1.5;
xlim([0 105])
ylabel('y_t')
title('Filtrovn')
subplot(3, 1, 3)
p = plot(t, y, 'k', t, y_smoothed, 'b');
p(2).LineWidth = 1.5;
xlim([0 105])
xlabel('t')
title('Vyrovnvn')

